package fos

import scala.util.parsing.combinator.syntactical.StandardTokenParsers
import scala.util.parsing.input._
import fos.Untyped._
import fos.Term._

/** This object implements a parser and runner for Project 2.
 */
object Main extends StandardTokenParsers {
  lexical.delimiters ++= List("(", ")", "\\", ".")
  import lexical.Identifier

  /** t ::= x
          | '\' x '.' t
          | t t
          | '(' t ')'
   */
  def term: Parser[Term] =
    ( AbsOrVar ~ rep(AbsOrVar) ^^ { case t1 ~ ts => (t1 :: ts).reduceLeft[Term]((t1, t2) => App(t1, t2)) }
    | failure("illegal start of term"))

  def AbsOrVar: Parser[Term] =
    ( ident ^^ { case chars => Var(chars) }
    | "\\" ~ ident ~ "." ~ term ^^ { case "\\" ~ chars ~ "." ~ t => Abs(chars, t) }
    | "(" ~ term ~ ")" ^^ { case "(" ~ (t: Term) ~ ")" => t }
    | failure("illegal start of .."))

  /** Returns a stream of terms, each being one step of reduction.
   *
   *  @param t      the initial term
   *  @param reduce the method that reduces a term by one step.
   *  @return       the stream of terms representing the big reduction.
   */
  def path(t: Term, reduce: Term => Term): LazyList[Term] =
    try {
      var t1 = reduce(t)
      LazyList.cons(t, path(t1, reduce))
    } catch {
      case NoReductionPossible(_) =>
        LazyList.cons(t, LazyList.empty)
    }

  def main(args: Array[String]): Unit = {
    val stdin = new java.io.BufferedReader(new java.io.InputStreamReader(System.in))
    val tokens = new lexical.Scanner(stdin.readLine())
    phrase(term)(tokens) match {
      case Success(trees, _) =>
        println("normal order: ")
        for (t <- path(trees, reduceNormalOrder))
          println(t)
        println("call-by-value: ")
        for (t <- path(trees, reduceCallByValue))
          println(t)

      case e =>
        println(e)
    }
  }
}
